/**
 * Endpoint for bulk destroy
 * @augments Endpoint
 * @class BulkRelatedDestroyEndpoint
 */
'use strict'

const Endpoint = require('../endpoint')

const { toType, defineRelatedOne, defineRelatedDestroy } = require('apemanmodel')
const validateDataType = require('../../validating/validate_data_type')
const validateModelExists = require('../../validating/validate_model_exists')
const validateDataGiven = require('../../validating/validate_data_given')

const co = require('co')

/** @lends BulkRelatedDestroyEndpoint */
class BulkRelatedDestroyEndpoint extends Endpoint {
  /**
   * @override
   */
  getDefaultConfig () {
    return {
      name: 'bulkRelatedDestroy'
    }
  }

  /**
   * @override
   */
  handle (ctx, next) {
    const s = this
    let succeed = s.bindSucceed(ctx)
    let fail = s.bindFail(ctx)

    let { state } = ctx
    let { id, data } = state
    return co(function * () {
      data = [].concat(data || [])

      return co(function * () {
        let error = s.validate(data, state)

        if (error) {
          return fail(error)
        }

        let count = yield s.bulkRelatedDestroy(id, data, state)
        return succeed({ count })
      })
    })
  }

  /**
   * @override
   */
  validate (data, state) {
    let { model, id, instance, association } = state
    let type = toType(association.target)
    let dataGivenError = validateDataGiven.bulk(data)
    if (dataGivenError) {
      return dataGivenError
    }

    let existsError = validateModelExists(model, id, instance)
    if (existsError) {
      return existsError
    }

    let typeError = validateDataType.bulk(data, type)
    if (typeError) {
      return typeError
    }

    return null
  }

  bulkRelatedDestroy (id, data, state) {
    let { model, association, transaction } = state
    let relatedOne = defineRelatedOne(model, association, transaction)
    let relatedDestroy = defineRelatedDestroy(model, association, transaction)
    return co(function * () {
      let count = 0
      for (let relatedId of data.map((data) => data.id)) {
        let related = yield relatedOne(id, relatedId)
        if (related) {
          yield relatedDestroy(related)
          count += 1
        }
      }
      return count
    })
  }
}

module.exports = BulkRelatedDestroyEndpoint
