/**
 * Endpoint for bulk create
 * @augments Endpoint
 * @class BulkRelationCreateEndpoint
 */
'use strict'

const Endpoint = require('../endpoint')

const { toType, toData, defineRelationCreate } = require('apemanmodel')
const validateDataType = require('../../validating/validate_data_type')
const validateModelExists = require('../../validating/validate_model_exists')
const validateDataGiven = require('../../validating/validate_data_given')

const co = require('co')

/** @lends BulkRelationCreateEndpoint */
class BulkRelationCreateEndpoint extends Endpoint {
  /**
   * @override
   */
  getDefaultConfig () {
    return {
      name: 'bulkRelationCreate'
    }
  }

  /**
   * @override
   */
  handle (ctx, next) {
    const s = this
    let succeed = s.bindSucceed(ctx)
    let fail = s.bindFail(ctx)

    let { state } = ctx
    let { id, data, association } = state
    return co(function * () {
      let error = s.validate(data, state)
      if (error) {
        return fail(error)
      }

      let created = yield s.bulkRelationCreate(id, data, state)

      return succeed(
        {},
        toData.withoutAttributes(association.target, created, {})
      )
    })
  }

  /**
   * @override
   */
  validate (data, state) {
    let { model, id, association, instance } = state
    let type = toType(association.target)

    let dataGivenError = validateDataGiven.bulk(data)
    if (dataGivenError) {
      return dataGivenError
    }

    let relatedTypeError = validateDataType.bulk(data, type)
    if (relatedTypeError) {
      return relatedTypeError
    }

    let existsError = validateModelExists(model, id, instance)
    if (existsError) {
      return existsError
    }

    return null
  }

  bulkRelationCreate (id, data, state) {
    let { model, association, transaction } = state
    let relationCreate = defineRelationCreate(model, association, transaction)
    return co(function * () {
      return yield data.map((data) => relationCreate(id, data))
    })
  }
}

module.exports = BulkRelationCreateEndpoint
